@@ -6,7 +6,9 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#define IP_VS_VERSION_CODE            0x000903
+#include <asm/types.h>          /* For __uXX types */
+
+#define IP_VS_VERSION_CODE            0x000904
 #define NVERSION(version)                       \
 	(version >> 16) & 0xFF,                 \
 	(version >> 8) & 0xFF,                  \
@@ -45,7 +47,7 @@
 /*
  *      IPVS socket options
  */
-#define IP_VS_BASE_CTL	(64+1024+64)            /* base */
+#define IP_VS_BASE_CTL		(64+1024+64)		/* base */
 
 #define IP_VS_SO_SET_NONE	IP_VS_BASE_CTL	        /* just peek */
 #define IP_VS_SO_SET_INSERT	(IP_VS_BASE_CTL+1)
@@ -59,13 +61,20 @@
 #define IP_VS_SO_SET_EDITDEST	(IP_VS_BASE_CTL+9)
 #define IP_VS_SO_SET_TIMEOUTS	(IP_VS_BASE_CTL+10)
 #define IP_VS_SO_SET_STARTDAEMON (IP_VS_BASE_CTL+11)
-#define IP_VS_SO_SET_STOPDAEMON  (IP_VS_BASE_CTL+12)
-#define IP_VS_SO_SET_MAX	IP_VS_SO_SET_STOPDAEMON
+#define IP_VS_SO_SET_STOPDAEMON (IP_VS_BASE_CTL+12)
+#define IP_VS_SO_SET_RESTORE    (IP_VS_BASE_CTL+13)
+#define IP_VS_SO_SET_SAVE       (IP_VS_BASE_CTL+14)
+#define IP_VS_SO_SET_MAX	IP_VS_SO_SET_SAVE
 
 #define IP_VS_SO_GET_VERSION	IP_VS_BASE_CTL
-#define IP_VS_SO_GET_SERVICE	(IP_VS_BASE_CTL+1)
-#define IP_VS_SO_GET_DESTS	(IP_VS_BASE_CTL+2)
-#define IP_VS_SO_GET_MAX	IP_VS_SO_GET_DESTS
+#define IP_VS_SO_GET_INFO	(IP_VS_BASE_CTL+1)
+#define IP_VS_SO_GET_SERVICES	(IP_VS_BASE_CTL+2)
+#define IP_VS_SO_GET_SERVICE	(IP_VS_BASE_CTL+3)
+#define IP_VS_SO_GET_DESTS	(IP_VS_BASE_CTL+4)
+#define IP_VS_SO_GET_DEST	(IP_VS_BASE_CTL+5)	/* not used now */
+#define IP_VS_SO_GET_TIMEOUTS	(IP_VS_BASE_CTL+6)
+#define IP_VS_SO_GET_DAEMON	(IP_VS_BASE_CTL+7)
+#define IP_VS_SO_GET_MAX	IP_VS_SO_GET_DAEMON
 
 
 /*
@@ -118,8 +127,39 @@ struct ip_vs_rule_user {
 };
 
 
+/*
+ *	IPVS statistics object (for user space)
+ */
+struct ip_vs_stats_user
+{
+	__u32                   conns;          /* connections scheduled */
+	__u32                   inpkts;         /* incoming packets */
+	__u32                   outpkts;        /* outgoing packets */
+	__u64                   inbytes;        /* incoming bytes */
+	__u64                   outbytes;       /* outgoing bytes */
+
+	__u32			cps;		/* current connection rate */
+	__u32			inpps;		/* current in packet rate */
+	__u32			outpps;		/* current out packet rate */
+	__u32			inbps;		/* current in byte rate */
+	__u32			outbps;		/* current out byte rate */
+};
+
+
+/* The argument to IP_VS_SO_GET_INFO */
+struct ip_vs_getinfo {
+	/* version number */
+	unsigned int	version;
+
+	/* size of connection hash table */
+	unsigned int	size;
+
+	/* number of virtual services */
+	unsigned int	num_services;
+};
+
 /* The argument to IP_VS_SO_GET_SERVICE */
-struct ip_vs_get_service {
+struct ip_vs_service_user {
 	/* which service: user fills this in */
 	u_int16_t	protocol;
 	u_int32_t	addr;           /* virtual address */
@@ -134,15 +174,21 @@ struct ip_vs_get_service {
 
 	/* number of real servers */
 	unsigned int    num_dests;
+
+	/* statistics */
+	struct ip_vs_stats_user stats;
 };
 
-struct ip_vs_dest_entry {
+struct ip_vs_dest_user {
 	u_int32_t	addr;           /* destination address */
 	u_int16_t	port;
 	unsigned	flags;		/* destination flags */
 	int		weight;         /* destination weight */
 	u_int16_t	activeconns;	/* active connections */
 	u_int16_t	inactconns;	/* inactive connections */
+
+	/* statistics */
+	struct ip_vs_stats_user stats;
 };
 
 /* The argument to IP_VS_SO_GET_DESTS */
@@ -157,7 +203,29 @@ struct ip_vs_get_dests {
 	unsigned int    num_dests;
 
 	/* the real servers */
-	struct ip_vs_dest_entry entrytable[0];
+	struct ip_vs_dest_user entrytable[0];
+};
+
+/* The argument to IP_VS_SO_GET_SERVICES */
+struct ip_vs_get_services {
+	/* number of virtual services */
+	unsigned int num_services;
+
+	/* service table */
+	struct ip_vs_service_user entrytable[0];
+};
+
+/* The argument to IP_VS_SO_GET_TIMEOUTS */
+struct ip_vs_timeout_user {
+	int             tcp_timeout;
+	int             tcp_fin_timeout;
+	int             udp_timeout;
+};
+
+/* The argument to IP_VS_SO_GET_DAEMON */
+struct ip_vs_daemon_user {
+	int	state;				/* sync daemon state */
+	char	mcast_ifn[IP_VS_IFNAME_MAXLEN];	/* multicast interface name */
 };
 
 
@@ -172,6 +240,8 @@ struct ip_vs_get_dests {
 #include <linux/netdevice.h>		/* for struct neighbour; */
 #include <net/dst.h>			/* for struct dst_entry */
 #include <net/route.h>			/* for ip_route_output */
+#include <net/tcp.h>
+#include <net/udp.h>
 
 
 #ifdef CONFIG_IP_VS_DEBUG
@@ -305,12 +375,19 @@ struct ip_vs_seq {
  */
 struct ip_vs_stats
 {
-	spinlock_t              lock;           /* spin lock */
 	__u32                   conns;          /* connections scheduled */
 	__u32                   inpkts;         /* incoming packets */
 	__u32                   outpkts;        /* outgoing packets */
 	__u64                   inbytes;        /* incoming bytes */
 	__u64                   outbytes;       /* outgoing bytes */
+
+	__u32			cps;		/* current connection rate */
+	__u32			inpps;		/* current in packet rate */
+	__u32			outpps;		/* current out packet rate */
+	__u32			inbps;		/* current in byte rate */
+	__u32			outbps;		/* current out byte rate */
+
+	spinlock_t              lock;           /* spin lock */
 };
 
 
@@ -640,11 +717,19 @@ extern void ip_vs_control_cleanup(void);
  *      (from ip_vs_sync.c)
  */
 extern int ip_vs_sync_state;
+extern char mcast_ifn[IP_VS_IFNAME_MAXLEN];
 extern int start_sync_thread(struct ip_vs_rule_user *u);
 extern int stop_sync_thread(void);
 extern void ip_vs_sync_conn(struct ip_vs_conn *cp);
 
 
+/*
+ *      IPVS rate estimator prototypes (from ip_vs_est.c)
+ */
+int ip_vs_new_estimator(struct ip_vs_stats *stats);
+void ip_vs_kill_estimator(struct ip_vs_stats *stats);
+
+
 /*
  *	This is a simple mechanism to ignore packets when
  *	we are loaded. Just set ip_vs_drop_rate to 'n' and
@@ -775,6 +860,52 @@ __ip_vs_get_out_rt(struct ip_vs_conn *cp, u32 rtos)
 	return rt;
 }
 
+static inline u16 ip_vs_check_diff(u32 old, u32 new, u16 oldsum)
+{
+	u32 diff[2] = { old, new };
+
+	return csum_fold(csum_partial((char *) diff, sizeof(diff),
+				      oldsum ^ 0xFFFF));
+}
+
+static inline void ip_vs_fast_check_update(union ip_vs_tphdr *h,
+	u32 oldip, u32 newip, u16 oldport, u16 newport, u8 protocol)
+{
+	u16 *checkp;
+
+	if (protocol == IPPROTO_TCP)
+		checkp = &h->th->check;
+	else
+		checkp = &h->uh->check;
+	*checkp = ip_vs_check_diff(~oldip, newip,
+		ip_vs_check_diff(oldport ^ 0xFFFF, newport, *checkp));
+	if (!*checkp && protocol == IPPROTO_UDP)
+		*checkp = 0xFFFF;
+}
+
+/*
+ *	Own skb_cow() beast, tweaked for rewriting commonly
+ *	used pointers in the code
+ */
+static inline struct sk_buff *vs_skb_cow(struct sk_buff **skb_p,
+				struct iphdr **iph_p, unsigned char **t_p)
+{
+	struct sk_buff *skb=(*skb_p);
+	if (skb_cloned(skb)) {
+		skb = skb_copy(skb, GFP_ATOMIC);
+		if (skb) {
+			/*
+			 *	skb changed, update other pointers
+			 */
+			struct iphdr *iph = skb->nh.iph;
+			kfree_skb(*skb_p);
+			*skb_p = skb;
+			*iph_p = iph;
+			*t_p = (char*) iph + iph->ihl * 4;
+		}
+	}
+	return skb;
+}
 
 #endif /* __KERNEL__ */
 
