@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.19 2001/08/27 02:40:54 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.20 2001/09/18 15:42:53 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -672,13 +672,15 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	EnterFunction(10);
 
 	h.raw = (char*) iph + iph->ihl * 4;
+	if ((skb = vs_skb_cow(&skb, &iph, &h.raw)) == NULL)
+		return NF_DROP;
 	size = ntohs(iph->tot_len) - (iph->ihl * 4);
 	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
 
 	/*
 	 *  Must do TCP/UDP checksum checking before mangling the packet
 	 */
-	if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
+	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
 		switch (skb->ip_summed) {
 		case CHECKSUM_NONE:
 			csum = csum_partial(h.raw + doff, size - doff, 0);
@@ -730,15 +732,21 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	dst_release(skb->dst);
 	skb->dst = &rt->u.dst;
 
-	/* mangle the packet */
-	iph->daddr = cp->daddr;
-	h.portp[1] = cp->dport;
-
 	/*
 	 *	Invalidate csum saving if it has ip_vs_app helper
 	 */
-	if (cp->app)
-		csum_ok = 0;
+	csum_ok = !cp->app;
+	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
+		/* Only port and addr are changed, make fast csum update */
+		ip_vs_fast_check_update(&h, iph->daddr, cp->daddr,
+			h.portp[1], cp->dport, iph->protocol);
+		if (skb->ip_summed == CHECKSUM_HW)
+			skb->ip_summed = CHECKSUM_NONE;
+	}
+
+	/* mangle the packet */
+	iph->daddr = cp->daddr;
+	h.portp[1] = cp->dport;
 
 	/*
 	 *	Attempt ip_vs_app call.
@@ -748,7 +756,6 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 		/*
 		 *	skb has changed, update pointers.
 		 */
-		/*  skb = *skb_p; */
 		iph = skb->nh.iph;
 		h.raw = (char*) iph + iph->ihl*4;
 		size = ntohs(iph->tot_len) - (iph->ihl * 4);
@@ -760,12 +767,13 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	if (!csum_ok) {
 		/* Transport's payload partial checksum */
 		csum = csum_partial(h.raw + doff, size - doff, 0);
+		skb->ip_summed = CHECKSUM_UNNECESSARY;
 	}
-	skb->csum = csum;
 
 	/*
 	 *	Protocol csum
 	 */
+	if (!csum_ok)
 	switch (iph->protocol) {
 	case IPPROTO_TCP:
 		h.th->check = 0;
