@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.21 2001/08/27 02:40:54 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.22 2001/09/18 15:42:54 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -189,32 +189,6 @@ ip_vs_conn_stats(struct ip_vs_conn *cp, struct ip_vs_service *svc)
 	spin_unlock(&ip_vs_stats.lock);
 }
 
-
-/*
- *	Own skb_cow() beast, tweaked for rewriting commonly
- *	used pointers in the code
- */
-static struct sk_buff *vs_skb_cow(struct sk_buff **skb_p,
-				  struct iphdr **iph_p, unsigned char **t_p)
-{
-	struct sk_buff *skb=(*skb_p);
-	if (skb_cloned(skb)) {
-		skb = skb_copy(skb, GFP_ATOMIC);
-		if (skb) {
-			/*
-			 *	skb changed, update other pointers
-			 */
-			struct iphdr *iph = skb->nh.iph;
-			kfree_skb(*skb_p);
-			*skb_p = skb;
-			*iph_p = iph;
-			*t_p = (char*) iph + iph->ihl * 4;
-		}
-	}
-	return skb;
-}
-
-
 /*
  *  IPVS persistent scheduling function
  *  It creates a connection entry according to its template if exists,
@@ -578,6 +552,7 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	__u16		*pptr;	/* port numbers from TCP/UDP contained header */
 	struct ip_vs_conn *cp;
 	unsigned short	 len   = ntohs(iph->tot_len) - (iph->ihl * 4);
+	unsigned short	clen, csize;
 
 	IP_VS_DBG(12, "outgoing ICMP (%d,%d) %u.%u.%u.%u->%u.%u.%u.%u\n",
 		  icmph->type, ntohs(icmp_id(icmph)),
@@ -596,18 +571,28 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 		return NF_ACCEPT;
 
 	/* Now find the contained IP header */
+	clen = len - sizeof(struct icmphdr);
+	if (clen < sizeof(struct iphdr))
+		return NF_DROP;
 	ciph = (struct iphdr *) (icmph + 1);
+	csize = ciph->ihl << 2;
+	if (clen < csize)
+		return NF_DROP;
 
 	/* We are only interested ICMPs generated from TCP or UDP packets */
 	if ((ciph->protocol != IPPROTO_UDP) && (ciph->protocol != IPPROTO_TCP))
 		return NF_ACCEPT;
 
+	/* We need at least TCP/UDP ports here */
+	if (clen < csize + sizeof(struct udphdr))
+		return NF_DROP;
+
 	/*
 	 * Find the ports involved - this packet was
 	 * incoming so the ports are right way round
 	 * (but reversed relative to outer IP header!)
 	 */
-	pptr = (__u16 *)&(((char *)ciph)[ciph->ihl*4]);
+	pptr = (__u16 *)&(((char *)ciph)[csize]);
 
 	/* Ensure the checksum is correct */
 	if (ip_compute_csum((unsigned char *) icmph, len)) {
@@ -648,6 +633,7 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	/* And finally the ICMP checksum */
 	icmph->checksum = 0;
 	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
+	skb->ip_summed = CHECKSUM_UNNECESSARY;
 
 	/* do the statistics and put it back */
 	ip_vs_out_stats(cp, skb);
@@ -742,7 +728,17 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	/* Check if the checksum is OK, when it is not UDP or check!=0 */
 	csum = 0;
 	csum_ok = 0;
-	if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
+	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
+		if (skb_is_nonlinear(skb)) {
+			if (skb_linearize(skb, GFP_ATOMIC) != 0) {
+				ip_vs_conn_put(cp);
+				return NF_DROP;
+			}
+			iph = skb->nh.iph;
+			h.raw = (char*) iph + iph->ihl * 4;
+			size = ntohs(iph->tot_len) - (iph->ihl * 4);
+			doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
+		}
 		switch (skb->ip_summed) {
 		case CHECKSUM_NONE:
 			csum = csum_partial(h.raw + doff, size - doff, 0);
@@ -776,18 +772,24 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	 */
 	size = skb->len - (h.raw - skb->nh.raw);
 
+	/*
+	 *	Invalidate csum saving if tunnel has connection helper
+	 */
+	csum_ok = !cp->app;
+	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
+		/* Only port and addr are changed, make fast csum update */
+		ip_vs_fast_check_update(&h, iph->saddr, cp->vaddr,
+			h.portp[0], cp->vport, iph->protocol);
+		if (skb->ip_summed == CHECKSUM_HW)
+			skb->ip_summed = CHECKSUM_NONE;
+	}
+
 	/*
 	 *	Set iph addr and port from ip_vs obj.
 	 */
 	iph->saddr = cp->vaddr;
 	h.portp[0] = cp->vport;
 
-	/*
-	 *	Invalidate csum saving if tunnel has connection helper
-	 */
-	if (cp->app)
-		csum_ok = 0;
-
 	/*
 	 *	Attempt ip_vs_app call.
 	 *	will fix ip_vs and iph seq stuff
@@ -809,8 +811,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	if (!csum_ok) {
 		/* transport's payload partial csum */
 		csum = csum_partial(h.raw + doff, size - doff, 0);
+		skb->ip_summed = CHECKSUM_UNNECESSARY;
 	}
-	skb->csum = csum;
 
 	IP_VS_DBG(11, "O-pkt: %s size=%d O-datacsum=%d\n",
 		  ip_vs_proto_name(iph->protocol), size, csum);
@@ -818,6 +820,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	/*
 	 *	Protocol checksum
 	 */
+	if (!csum_ok)
 	switch (iph->protocol) {
 	case IPPROTO_TCP:
 		h.th->check = 0;
@@ -888,15 +891,24 @@ unsigned int check_for_ip_vs_out(struct sk_buff **skb_p,
 static int ip_vs_in_icmp(struct sk_buff **skb_p)
 {
 	struct sk_buff	*skb   = *skb_p;
-	struct iphdr	*iph   = skb->nh.iph;
-	struct icmphdr  *icmph = (struct icmphdr *)((char *)iph+(iph->ihl<<2));
+	struct iphdr    *iph;
+	struct icmphdr  *icmph;
 	struct iphdr    *ciph;	/* The ip header contained within the ICMP */
 	__u16	        *pptr;	/* port numbers from TCP/UDP contained header */
-	unsigned short   len   = ntohs(iph->tot_len) - (iph->ihl * 4);
+	unsigned short   len;
+	unsigned short	clen, csize;
 	struct ip_vs_conn *cp;
 	struct rtable *rt;			/* Route to the other host */
 	int    mtu;
 
+	/* Sometimes we come here from ip_vs_forward_icmp */
+	if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0)
+		return NF_DROP;
+
+	iph = skb->nh.iph;
+	icmph = (struct icmphdr *)((char *)iph+(iph->ihl<<2));
+	len = ntohs(iph->tot_len) - (iph->ihl * 4);
+
 	IP_VS_DBG(12, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
 		  icmph->type, ntohs(icmp_id(icmph)),
 		  NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
@@ -910,13 +922,23 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	 * If we get here we have an ICMP error of one of the above 3 types
 	 * Now find the contained IP header
 	 */
+	clen = len - sizeof(struct icmphdr);
+	if (clen < sizeof(struct iphdr))
+		return NF_DROP;
 	ciph = (struct iphdr *) (icmph + 1);
+	csize = ciph->ihl << 2;
+	if (clen < csize)
+		return NF_DROP;
 
 	/* We are only interested ICMPs generated from TCP or UDP packets */
 	if ((ciph->protocol != IPPROTO_UDP) &&
 	    (ciph->protocol != IPPROTO_TCP))
 		return NF_ACCEPT;
 
+	/* We need at least TCP/UDP ports here */
+	if (clen < csize + sizeof(struct udphdr))
+		return NF_DROP;
+
 	/* Ensure the checksum is correct */
 	if (ip_compute_csum((unsigned char *) icmph, len)) {
 		/* Failed checksum! */
@@ -925,7 +947,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 	}
 
-	pptr = (__u16 *)&(((char *)ciph)[ciph->ihl*4]);
+	pptr = (__u16 *)&(((char *)ciph)[csize]);
 
 	IP_VS_DBG(11, "Handling incoming ICMP for "
 		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
@@ -981,6 +1003,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	/* And finally the ICMP checksum */
 	icmph->checksum = 0;
 	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
+	skb->ip_summed = CHECKSUM_UNNECESSARY;
 
 	IP_VS_DBG(11, "Forwarding incoming ICMP to "
 		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
@@ -1013,9 +1036,9 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	ip_vs_conn_put(cp);
 	return NF_STOLEN;
 
-  tx_error_icmp:
+ tx_error_icmp:
 	dst_link_failure(skb);
-  tx_error:
+ tx_error:
 	dev_kfree_skb(skb);
 	ip_vs_conn_put(cp);
 	return NF_STOLEN;
