@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_ctl.c,v 1.19 2001/08/27 02:40:54 wensong Exp $
+ * Version:     $Id: ip_vs_ctl.c,v 1.21 2001/09/19 01:14:24 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -67,6 +67,9 @@ int ip_vs_drop_rate = 0;
 int ip_vs_drop_counter = 0;
 atomic_t ip_vs_dropentry = ATOMIC_INIT(0);
 
+/* number of virtual services */
+static int ip_vs_num_services = 0;
+
 /* sysctl variables */
 static int sysctl_ip_vs_drop_entry = 0;
 static int sysctl_ip_vs_drop_packet = 0;
@@ -100,26 +103,26 @@ void update_defense_level(void)
 	spin_lock(&__ip_vs_dropentry_lock);
 	switch (sysctl_ip_vs_drop_entry) {
 	case 0:
-		atomic_set(&ip_vs_dropentry,0);
+		atomic_set(&ip_vs_dropentry, 0);
 		break;
 	case 1:
 		if (nomem) {
-			atomic_set(&ip_vs_dropentry,1);
+			atomic_set(&ip_vs_dropentry, 1);
 			sysctl_ip_vs_drop_entry = 2;
 		} else {
-			atomic_set(&ip_vs_dropentry,0);
+			atomic_set(&ip_vs_dropentry, 0);
 		}
 		break;
 	case 2:
 		if (nomem) {
-			atomic_set(&ip_vs_dropentry,1);
+			atomic_set(&ip_vs_dropentry, 1);
 		} else {
-			atomic_set(&ip_vs_dropentry,0);
+			atomic_set(&ip_vs_dropentry, 0);
 			sysctl_ip_vs_drop_entry = 1;
 		};
 		break;
 	case 3:
-		atomic_set(&ip_vs_dropentry,1);
+		atomic_set(&ip_vs_dropentry, 1);
 		break;
 	}
 	spin_unlock(&__ip_vs_dropentry_lock);
@@ -636,29 +639,23 @@ static void ip_vs_trash_cleanup(void)
 /*
  *  Update a destination in the given service
  */
-static int __ip_vs_update_dest(struct ip_vs_service *svc,
+static void __ip_vs_update_dest(struct ip_vs_service *svc,
 				struct ip_vs_dest *dest,
 				struct ip_vs_rule_user *ur)
 {
 	int conn_flags;
-	unsigned atype;
-
-	atype = inet_addr_type(ur->daddr);
-	if (atype != RTN_LOCAL && atype != RTN_UNICAST)
-		return -EINVAL;
 
 	/*
 	 *    Set the weight and the flags
 	 */
 	atomic_set(&dest->weight, ur->weight);
-	conn_flags = ur->conn_flags;
 
-	conn_flags |= IP_VS_CONN_F_INACTIVE;
+	conn_flags = ur->conn_flags | IP_VS_CONN_F_INACTIVE;
 
 	/*
 	 *    Check if local node and update the flags
 	 */
-	if (atype == RTN_LOCAL) {
+	if (inet_addr_type(ur->daddr) == RTN_LOCAL) {
 		conn_flags = (conn_flags & ~IP_VS_CONN_F_FWD_MASK)
 			| IP_VS_CONN_F_LOCALNODE;
 	}
@@ -668,9 +665,7 @@ static int __ip_vs_update_dest(struct ip_vs_service *svc,
 	 */
 	if ((conn_flags & IP_VS_CONN_F_FWD_MASK) != 0) {
 		conn_flags |= IP_VS_CONN_F_NOOUTPUT;
-		atomic_set(&dest->conn_flags, conn_flags);
 	} else {
-		atomic_set(&dest->conn_flags, conn_flags);
 		/*
 		 *    Put the real service in ip_vs_rtable if not present.
 		 *    For now only for NAT!
@@ -679,6 +674,7 @@ static int __ip_vs_update_dest(struct ip_vs_service *svc,
 		ip_vs_rs_hash(dest);
 		write_unlock_bh(&__ip_vs_rs_lock);
 	}
+	atomic_set(&dest->conn_flags, conn_flags);
 
 	/* bind the service */
 	if (!dest->svc) {
@@ -692,8 +688,6 @@ static int __ip_vs_update_dest(struct ip_vs_service *svc,
 
 	/* set the dest status flags */
 	dest->flags |= IP_VS_DEST_F_AVAILABLE;
-
-	return 0;
 }
 
 
@@ -702,15 +696,19 @@ static int __ip_vs_update_dest(struct ip_vs_service *svc,
  */
 static int
 ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
-	struct ip_vs_dest **destp)
+	       struct ip_vs_dest **destp)
 {
 	struct ip_vs_dest *dest;
-	int ret;
+	unsigned atype;
 
 	EnterFunction(2);
 
-	*destp = dest = (struct ip_vs_dest*) kmalloc(sizeof(struct ip_vs_dest),
-					GFP_ATOMIC);
+	atype = inet_addr_type(ur->daddr);
+	if (atype != RTN_LOCAL && atype != RTN_UNICAST)
+		return -EINVAL;
+
+	*destp = dest = (struct ip_vs_dest*)
+		kmalloc(sizeof(struct ip_vs_dest), GFP_ATOMIC);
 	if (dest == NULL) {
 		IP_VS_ERR("ip_vs_new_dest: kmalloc failed.\n");
 		return -ENOMEM;
@@ -731,12 +729,11 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
 	INIT_LIST_HEAD(&dest->d_list);
 	dest->dst_lock = SPIN_LOCK_UNLOCKED;
 	dest->stats.lock = SPIN_LOCK_UNLOCKED;
-	ret = __ip_vs_update_dest(svc, dest, ur);
-	if (ret) kfree(dest);
+	__ip_vs_update_dest(svc, dest, ur);
+	ip_vs_new_estimator(&dest->stats);
 
 	LeaveFunction(2);
-
-	return ret;
+	return 0;
 }
 
 
@@ -780,16 +777,15 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 			  dest->vfwmark,
 			  NIPQUAD(dest->vaddr),
 			  ntohs(dest->vport));
-
-		ret = __ip_vs_update_dest(svc, dest, ur);
-		if (ret)
-			return ret;
+		__ip_vs_update_dest(svc, dest, ur);
 
 		/*
 		 * Get the destination from the trash
 		 */
 		list_del(&dest->n_list);
 
+		ip_vs_new_estimator(&dest->stats);
+
 		write_lock_bh(&__ip_vs_svc_lock);
 
 		/*
@@ -850,7 +846,6 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 	struct ip_vs_dest *dest;
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
-	int ret;
 
 	EnterFunction(2);
 
@@ -868,11 +863,11 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 		return -ENOENT;
 	}
 
-	ret = __ip_vs_update_dest(svc, dest, ur);
+	__ip_vs_update_dest(svc, dest, ur);
 
 	LeaveFunction(2);
 
-	return ret;
+	return 0;
 }
 
 
@@ -882,6 +877,8 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 
 static void __ip_vs_del_dest(struct ip_vs_dest *dest)
 {
+	ip_vs_kill_estimator(&dest->stats);
+
 	/*
 	 *  Remove it from the d-linked list with the real services.
 	 */
@@ -984,14 +981,9 @@ static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 static int
 ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 {
-	__u16  protocol = ur->protocol;
-	__u32 vaddr = ur->vaddr;
-	__u16 vport = ur->vport;
-	__u32 vfwmark = ur->vfwmark;
-
 	int ret = 0;
 	struct ip_vs_scheduler *sched;
-	struct ip_vs_service *svc;
+	struct ip_vs_service *svc = NULL;
 
 	MOD_INC_USE_COUNT;
 
@@ -1002,24 +994,23 @@ ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 	if (sched == NULL) {
 		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
 			   ur->sched_name);
-		MOD_DEC_USE_COUNT;
-		return -ENOENT;
+		ret = -ENOENT;
+		goto out_err;
 	}
 
 	svc = (struct ip_vs_service*)
 		kmalloc(sizeof(struct ip_vs_service), GFP_ATOMIC);
 	if (svc == NULL) {
-		MOD_DEC_USE_COUNT;
 		IP_VS_DBG(1, "ip_vs_add_service: kmalloc failed.\n");
 		ret = -ENOMEM;
 		goto out_err;
 	}
 	memset(svc, 0, sizeof(struct ip_vs_service));
 
-	svc->protocol = protocol;
-	svc->addr = vaddr;
-	svc->port = vport;
-	svc->fwmark = vfwmark;
+	svc->protocol = ur->protocol;
+	svc->addr = ur->vaddr;
+	svc->port = ur->vport;
+	svc->fwmark = ur->vfwmark;
 	svc->flags = ur->vs_flags;
 	svc->timeout = ur->timeout;
 	svc->netmask = ur->netmask;
@@ -1039,9 +1030,9 @@ ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 	/*
 	 *    Update the virtual service counters
 	 */
-	if (vport == FTPPORT)
+	if (svc->port == FTPPORT)
 		atomic_inc(&ip_vs_ftpsvc_counter);
-	else if (vport == 0)
+	else if (svc->port == 0)
 		atomic_inc(&ip_vs_nullsvc_counter);
 
 	/*
@@ -1057,16 +1048,18 @@ ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 	ip_vs_svc_hash(svc);
 	write_unlock_bh(&__ip_vs_svc_lock);
 
+	ip_vs_new_estimator(&svc->stats);
+	ip_vs_num_services++;
 	*svc_p = svc;
 	return 0;
 
   out_err:
 	if (sched->module)
 		__MOD_DEC_USE_COUNT(sched->module);
-	MOD_DEC_USE_COUNT;
 	if (svc) {
 		kfree(svc);
 	}
+	MOD_DEC_USE_COUNT;
 	return ret;
 }
 
@@ -1147,12 +1140,15 @@ static int ip_vs_edit_service(struct ip_vs_service *svc,
  *  Delete a service from the service list
  *  The service must be unlinked, unlocked and not referenced!
  */
-static int __ip_vs_del_service(struct ip_vs_service *svc)
+static void __ip_vs_del_service(struct ip_vs_service *svc)
 {
 	struct list_head *l;
 	struct ip_vs_dest *dest;
 	struct ip_vs_scheduler *old_sched;
 
+	ip_vs_num_services--;
+	ip_vs_kill_estimator(&svc->stats);
+
 	/*
 	 *    Unbind scheduler
 	 */
@@ -1185,7 +1181,6 @@ static int __ip_vs_del_service(struct ip_vs_service *svc)
 	if (atomic_read(&svc->refcnt) == 0)
 		kfree(svc);
 	MOD_DEC_USE_COUNT;
-	return 0;
 }
 
 /*
@@ -1193,8 +1188,6 @@ static int __ip_vs_del_service(struct ip_vs_service *svc)
  */
 static int ip_vs_del_service(struct ip_vs_service *svc)
 {
-	int ret;
-
 	if (svc == NULL)
 		return -EEXIST;
 
@@ -1210,11 +1203,11 @@ static int ip_vs_del_service(struct ip_vs_service *svc)
 	 */
 	while (atomic_read(&svc->usecnt) > 1) {};
 
-	ret = __ip_vs_del_service(svc);
+	__ip_vs_del_service(svc);
 
 	write_unlock_bh(&__ip_vs_svc_lock);
 
-	return ret;
+	return 0;
 }
 
 
@@ -1534,171 +1527,47 @@ static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
 }
 
 
-struct ip_vs_stats ip_vs_stats = {SPIN_LOCK_UNLOCKED, 0, 0};
+struct ip_vs_stats ip_vs_stats;
 
 static int
 ip_vs_stats_get_info(char *buf, char **start, off_t offset, int length)
 {
-	int idx;
 	int len=0;
 	off_t pos=0;
-	char temp[128];
-	struct ip_vs_service *svc;
-	struct ip_vs_dest *dest;
-	struct list_head *l, *e, *p, *q;
+	char temp[64];
 
-	pos += 128;
+	pos += 320;
 	if (pos > offset) {
-		len += sprintf(buf+len, "%-63s\n",
-/*                                01234567 01234567 01234567 0123456701234567 0123456701234567 */
-			       "TotalConns   InPkts  OutPkts          InBytes         OutBytes");
+		len += sprintf(buf+len, "%-63s\n%-63s\n",
+/*                              01234567 01234567 01234567 0123456701234567 0123456701234567 */
+			       "   Total Incoming Outgoing         Incoming         Outgoing",
+			       "   Conns  Packets  Packets            Bytes            Bytes");
+
 		spin_lock(&ip_vs_stats.lock);
-		sprintf(temp, "  %8X %8X %8X %8X%08X %8X%08X",
+		sprintf(temp, "%8X %8X %8X %8X%08X %8X%08X",
 			ip_vs_stats.conns,
 			ip_vs_stats.inpkts,
 			ip_vs_stats.outpkts,
-			(__u32)(ip_vs_stats.inbytes >> 32),
+			(__u32)(ip_vs_stats.inbytes>>32),
 			(__u32)ip_vs_stats.inbytes,
-			(__u32)(ip_vs_stats.outbytes >> 32),
+			(__u32)(ip_vs_stats.outbytes>>32),
 			(__u32)ip_vs_stats.outbytes);
-		spin_unlock(&ip_vs_stats.lock);
+		len += sprintf(buf+len, "%-62s\n\n", temp);
+
+		len += sprintf(buf+len, "%-63s\n",
+/*                              01234567 01234567 01234567 0123456701234567 0123456701234567 */
+			       " Conns/s   Pkts/s   Pkts/s          Bytes/s          Bytes/s");
+		sprintf(temp, "%8X %8X %8X %16X %16X",
+			ip_vs_stats.cps,
+			ip_vs_stats.inpps,
+			ip_vs_stats.outpps,
+			ip_vs_stats.inbps,
+			ip_vs_stats.outbps);
 		len += sprintf(buf+len, "%-63s\n", temp);
-	}
 
-	read_lock_bh(&__ip_vs_svc_lock);
-
-	/* print the service statistics */
-	pos += 128;
-	if (pos > offset) {
-		len += sprintf(buf+len, "%-127s\n",
-			       "\nVirtual Service\n"
-			       "Pro VirtService      Conns   InPkts  OutPkts          InBytes         OutBytes");
+		spin_unlock(&ip_vs_stats.lock);
 	}
 
-	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
-		l = &ip_vs_svc_table[idx];
-		for (e=l->next; e!=l; e=e->next) {
-			svc = list_entry(e, struct ip_vs_service, s_list);
-			pos += 128;
-			if (pos <= offset)
-				continue;
-			spin_lock(&svc->stats.lock);
-			sprintf(temp, "%3s %08X:%04X %8X %8X %8X %8X%08X %8X%08X",
-				ip_vs_proto_name(svc->protocol),
-				ntohl(svc->addr),
-				ntohs(svc->port),
-				svc->stats.conns,
-				svc->stats.inpkts,
-				svc->stats.outpkts,
-				(__u32)(svc->stats.inbytes >> 32),
-				(__u32)svc->stats.inbytes,
-				(__u32)(svc->stats.outbytes >> 32),
-				(__u32)svc->stats.outbytes);
-			spin_unlock(&svc->stats.lock);
-			len += sprintf(buf+len, "%-127s\n", temp);
-			if (pos >= offset+length)
-				goto done;
-		}
-	}
-
-	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
-		l = &ip_vs_svc_fwm_table[idx];
-		for (e=l->next; e!=l; e=e->next) {
-			svc = list_entry(e, struct ip_vs_service, f_list);
-			pos += 128;
-			if (pos <= offset)
-				continue;
-			spin_lock(&svc->stats.lock);
-			sprintf(temp, "FWM %08X      %8X %8X %8X %8X%08X %8X%08X",
-				svc->fwmark,
-				svc->stats.conns,
-				svc->stats.inpkts,
-				svc->stats.outpkts,
-				(__u32)(svc->stats.inbytes >> 32),
-				(__u32)svc->stats.inbytes,
-				(__u32)(svc->stats.outbytes >> 32),
-				(__u32)svc->stats.outbytes);
-			spin_unlock(&svc->stats.lock);
-			len += sprintf(buf+len, "%-127s\n", temp);
-			if (pos >= offset+length)
-				goto done;
-		}
-	}
-
-	/* print the real server statistics */
-	pos += 128;
-	if (pos > offset) {
-		len += sprintf(buf+len, "%-127s\n",
-			       "\nReal Service\n"
-			       "Pro VirtService   RealService      Conns   InPkts  OutPkts          InBytes         OutBytes");
-	}
-
-	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
-		l = &ip_vs_svc_table[idx];
-		for (e=l->next; e!=l; e=e->next) {
-			svc = list_entry(e, struct ip_vs_service, s_list);
-			p = &svc->destinations;
-			for (q=p->next; q!=p; q=q->next) {
-				dest = list_entry(q, struct ip_vs_dest, n_list);
-				pos += 128;
-				if (pos <= offset)
-					continue;
-				spin_lock(&dest->stats.lock);
-				sprintf(temp,
-					"%3s %08X:%04X %08X:%04X %8X %8X %8X %8X%08X %8X%08X",
-					ip_vs_proto_name(svc->protocol),
-					ntohl(svc->addr),
-					ntohs(svc->port),
-					ntohl(dest->addr),
-					ntohs(dest->port),
-					dest->stats.conns,
-					dest->stats.inpkts,
-					dest->stats.outpkts,
-					(__u32)(dest->stats.inbytes >> 32),
-					(__u32)dest->stats.inbytes,
-					(__u32)(dest->stats.outbytes >> 32),
-					(__u32)dest->stats.outbytes);
-				spin_unlock(&dest->stats.lock);
-				len += sprintf(buf+len, "%-127s\n", temp);
-				if (pos >= offset+length)
-					goto done;
-			}
-		}
-	}
-
-	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
-		l = &ip_vs_svc_fwm_table[idx];
-		for (e=l->next; e!=l; e=e->next) {
-			svc = list_entry(e, struct ip_vs_service, f_list);
-			p = &svc->destinations;
-			for (q=p->next; q!=p; q=q->next) {
-				dest = list_entry(q,struct ip_vs_dest,n_list);
-				pos += 128;
-				if (pos <= offset)
-					continue;
-				spin_lock(&dest->stats.lock);
-				sprintf(temp,
-					"FWM %08X      %08X:%04X %8X %8X %8X %8X%08X %8X%08X",
-					svc->fwmark,
-					ntohl(dest->addr),
-					ntohs(dest->port),
-					dest->stats.conns,
-					dest->stats.inpkts,
-					dest->stats.outpkts,
-					(__u32)(dest->stats.inbytes >> 32),
-					(__u32)dest->stats.inbytes,
-					(__u32)(dest->stats.outbytes >> 32),
-					(__u32)dest->stats.outbytes);
-				spin_unlock(&dest->stats.lock);
-				len += sprintf(buf+len, "%-127s\n", temp);
-				if (pos >= offset+length)
-					goto done;
-			}
-		}
-	}
-  done:
-	read_unlock_bh(&__ip_vs_svc_lock);
-
 	*start = buf+len-(pos-offset);          /* Start of wanted data */
 	len = pos-offset;
 	if (len > length)
@@ -1859,6 +1728,77 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 }
 
 
+static inline void
+__ip_vs_copy_stats(struct ip_vs_stats_user *dst, struct ip_vs_stats *src)
+{
+	spin_lock(&src->lock);
+	memcpy(dst, src, (char*)&src->lock - (char*)src);
+	spin_unlock(&src->lock);
+}
+
+static inline int
+__ip_vs_get_service_entries(const struct ip_vs_get_services *get,
+			    struct ip_vs_get_services *uptr)
+{
+	int idx, count=0;
+	struct ip_vs_service *svc;
+	struct list_head *l;
+	struct ip_vs_service_user entry;
+	int ret = 0;
+
+	down_interruptible(&__ip_vs_mutex);
+	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+		list_for_each (l, &ip_vs_svc_table[idx]) {
+			if (count >= get->num_services)
+				goto out;
+			svc = list_entry(l, struct ip_vs_service, s_list);
+			entry.protocol = svc->protocol;
+			entry.addr = svc->addr;
+			entry.port = svc->port;
+			entry.fwmark = svc->fwmark;
+			strcpy(entry.sched_name, svc->scheduler->name);
+			entry.flags = svc->flags;
+			entry.timeout = svc->timeout;
+			entry.netmask = svc->netmask;
+			entry.num_dests = svc->num_dests;
+			__ip_vs_copy_stats(&entry.stats, &svc->stats);
+			if (copy_to_user(&uptr->entrytable[count],
+					 &entry, sizeof(entry))) {
+				ret = -EFAULT;
+				goto out;
+			}
+			count++;
+		}
+	}
+
+	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+		list_for_each (l, &ip_vs_svc_fwm_table[idx]) {
+			if (count >= get->num_services)
+				goto out;
+			svc = list_entry(l, struct ip_vs_service, f_list);
+			entry.protocol = svc->protocol;
+			entry.addr = svc->addr;
+			entry.port = svc->port;
+			entry.fwmark = svc->fwmark;
+			strcpy(entry.sched_name, svc->scheduler->name);
+			entry.flags = svc->flags;
+			entry.timeout = svc->timeout;
+			entry.netmask = svc->netmask;
+			entry.num_dests = svc->num_dests;
+			__ip_vs_copy_stats(&entry.stats, &svc->stats);
+			if (copy_to_user(&uptr->entrytable[count],
+					 &entry, sizeof(entry))) {
+				ret = -EFAULT;
+				goto out;
+			}
+			count++;
+		}
+	}
+ out:
+	up(&__ip_vs_mutex);
+	return ret;
+}
+
 static inline int
 __ip_vs_get_dest_entries(const struct ip_vs_get_dests *get,
 			 struct ip_vs_get_dests *uptr)
@@ -1873,13 +1813,15 @@ __ip_vs_get_dest_entries(const struct ip_vs_get_dests *get,
 		svc = __ip_vs_service_get(get->protocol,
 					  get->addr, get->port);
 	if (svc) {
-		int i = 0;
+		int count = 0;
 		struct ip_vs_dest *dest;
 		struct list_head *l, *e;
-		struct ip_vs_dest_entry entry;
+		struct ip_vs_dest_user entry;
 
 		l = &svc->destinations;
 		for (e=l->next; e!=l; e=e->next) {
+			if (count >= get->num_dests)
+				break;
 			dest = list_entry(e, struct ip_vs_dest, n_list);
 			entry.addr = dest->addr;
 			entry.port = dest->port;
@@ -1887,17 +1829,17 @@ __ip_vs_get_dest_entries(const struct ip_vs_get_dests *get,
 			entry.weight = atomic_read(&dest->weight);
 			entry.activeconns = atomic_read(&dest->activeconns);
 			entry.inactconns = atomic_read(&dest->inactconns);
-			if (copy_to_user(&uptr->entrytable[i],
+			__ip_vs_copy_stats(&entry.stats, &dest->stats);
+			if (copy_to_user(&uptr->entrytable[count],
 					 &entry, sizeof(entry))) {
 				ret = -EFAULT;
 				break;
 			}
-			if (++i >= get->num_dests)
-				break;
+			count++;
 		}
 		ip_vs_service_put(svc);
 	} else
-		ret = -EEXIST;
+		ret = -ESRCH;
 	up(&__ip_vs_mutex);
 	return ret;
 }
@@ -1914,20 +1856,51 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 	switch (cmd) {
 	case IP_VS_SO_GET_VERSION:
 	{
-		char buf[32];
+		char buf[64];
 
-		sprintf(buf, "IPVS %d.%d.%d", NVERSION(IP_VS_VERSION_CODE));
-		if (*len < sizeof(buf)+1)
+		sprintf(buf, "IP Virtual Server version %d.%d.%d (size=%d)",
+			NVERSION(IP_VS_VERSION_CODE), IP_VS_CONN_TAB_SIZE);
+		if (*len < strlen(buf)+1)
 			return -EINVAL;
-		if (copy_to_user(user, buf, sizeof(buf)+1) != 0)
+		if (copy_to_user(user, buf, strlen(buf)+1) != 0)
 			return -EFAULT;
-		*len = sizeof(buf)+1;
+		*len = strlen(buf)+1;
+	}
+	break;
+
+	case IP_VS_SO_GET_INFO:
+	{
+		struct ip_vs_getinfo info;
+		info.version = IP_VS_VERSION_CODE;
+		info.size = IP_VS_CONN_TAB_SIZE;
+		info.num_services = ip_vs_num_services;
+		if (copy_to_user(user, &info, sizeof(info)) != 0)
+			return -EFAULT;
+	}
+	break;
+
+	case IP_VS_SO_GET_SERVICES:
+	{
+		struct ip_vs_get_services get;
+
+		if (*len < sizeof(get)) {
+			IP_VS_ERR("length: %u < %u\n", *len, sizeof(get));
+			return -EINVAL;
+		}
+		if (copy_from_user(&get, user, sizeof(get)))
+			return -EFAULT;
+		if (*len != (sizeof(get)+sizeof(struct ip_vs_service_user)*get.num_services)) {
+			IP_VS_ERR("length: %u != %u\n", *len,
+				  sizeof(get)+sizeof(struct ip_vs_service_user)*get.num_services);
+			return -EINVAL;
+		}
+		ret = __ip_vs_get_service_entries(&get, user);
 	}
 	break;
 
 	case IP_VS_SO_GET_SERVICE:
 	{
-		struct ip_vs_get_service get;
+		struct ip_vs_service_user get;
 		struct ip_vs_service *svc;
 
 		if (*len != sizeof(get)) {
@@ -1949,11 +1922,12 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 			get.timeout = svc->timeout;
 			get.netmask = svc->netmask;
 			get.num_dests = svc->num_dests;
+			__ip_vs_copy_stats(&get.stats, &svc->stats);
 			if (copy_to_user(user, &get, *len) != 0)
 				ret = -EFAULT;
 			ip_vs_service_put(svc);
 		} else
-			ret = -EEXIST;
+			ret = -ESRCH;
 		up(&__ip_vs_mutex);
 	}
 	break;
@@ -1969,15 +1943,46 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 		if (copy_from_user(&get, user, sizeof(get)))
 			return -EFAULT;
 		if (*len != (sizeof(get) +
-			     sizeof(struct ip_vs_dest_entry)*get.num_dests)) {
+			     sizeof(struct ip_vs_dest_user)*get.num_dests)) {
 			IP_VS_ERR("length: %u != %u\n", *len,
-				  sizeof(get)+sizeof(struct ip_vs_dest_entry)*get.num_dests);
+				  sizeof(get)+sizeof(struct ip_vs_dest_user)*get.num_dests);
 			return -EINVAL;
 		}
 		ret = __ip_vs_get_dest_entries(&get, user);
 	}
 	break;
 
+	case IP_VS_SO_GET_TIMEOUTS:
+	{
+		struct ip_vs_timeout_user u;
+
+		if (*len < sizeof(u)) {
+			IP_VS_ERR("length: %u < %u\n", *len, sizeof(u));
+			return -EINVAL;
+		}
+		u.tcp_timeout = vs_timeout_table.timeout[IP_VS_S_ESTABLISHED];
+		u.tcp_fin_timeout = vs_timeout_table.timeout[IP_VS_S_FIN_WAIT];
+		u.udp_timeout = vs_timeout_table.timeout[IP_VS_S_UDP];
+		if (copy_to_user(user, &u, sizeof(u)) != 0)
+			ret = -EFAULT;
+	}
+	break;
+
+	case IP_VS_SO_GET_DAEMON:
+	{
+		struct ip_vs_daemon_user u;
+
+		if (*len < sizeof(u)) {
+			IP_VS_ERR("length: %u < %u\n", *len, sizeof(u));
+			return -EINVAL;
+		}
+		u.state = ip_vs_sync_state;
+		strcpy(u.mcast_ifn, mcast_ifn);
+		if (copy_to_user(user, &u, sizeof(u)) != 0)
+			ret = -EFAULT;
+	}
+	break;
+
 	default:
 		ret = -EINVAL;
 	}
@@ -2023,6 +2028,10 @@ int __init ip_vs_control_init(void)
 		INIT_LIST_HEAD(&ip_vs_rtable[idx]);
 	}
 
+	memset(&ip_vs_stats, 0, sizeof(ip_vs_stats));
+	ip_vs_stats.lock = SPIN_LOCK_UNLOCKED;
+	ip_vs_new_estimator(&ip_vs_stats);
+
 	LeaveFunction(2);
 	return 0;
 }
@@ -2031,6 +2040,7 @@ void __exit ip_vs_control_cleanup(void)
 {
 	EnterFunction(2);
 	ip_vs_trash_cleanup();
+	ip_vs_kill_estimator(&ip_vs_stats);
 	unregister_sysctl_table(ipv4_vs_table.sysctl_header);
 	proc_net_remove("ip_vs_stats");
 	proc_net_remove("ip_vs");
