@@ -2,7 +2,7 @@
  *      ipvsadm - IP Virtual Server ADMinistration program
  *                for IPVS NetFilter Module in kernel 2.4
  *
- *      Version: $Id: ipvsadm.c,v 1.13 2001/08/27 02:40:55 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.14 2001/09/18 15:42:54 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *               Peter Kese <peter.kese@ijs.si>
@@ -103,15 +103,16 @@
 #include <sys/wait.h>           /* For waitpid */
 #include <arpa/inet.h>
 
-#include <asm/types.h>          /* For __uXX types */
 #include <net/if.h>
 #include <netinet/ip_icmp.h>
 #include <netinet/udp.h>
 #include <netinet/tcp.h>
 
+#ifdef HAVE_NET_IP_VS_H
+#include <net/ip_vs.h>
+#else
 #include "ip_vs.h"
-//#include <net/ip_vs.h>
-#include "config_stream.h"
+#endif
 
 #ifdef HAVE_POPT
 #include "popt.h"
@@ -120,37 +121,34 @@
 #define IPVS_OPTION_PROCESSING          "getopt_long"
 #endif
 
+#include "config_stream.h"
+#include "libipvs/libipvs.h"
+
 #define IPVSADM_VERSION_NO              "v" VERSION
-#define IPVSADM_VERSION_DATE            "2001/08/27"
+#define IPVSADM_VERSION_DATE            "2001/09/18"
 #define IPVSADM_VERSION         IPVSADM_VERSION_NO " " IPVSADM_VERSION_DATE
 
 #define MINIMUM_IPVS_VERSION_MAJOR      0
 #define MINIMUM_IPVS_VERSION_MINOR      9
-#define MINIMUM_IPVS_VERSION_PATCH      3
+#define MINIMUM_IPVS_VERSION_PATCH      4
 
 #ifndef IPVS_VERSION
 #define IPVS_VERSION(x,y,z) (((x)<<16)+((y)<<8)+(z))
 #endif
 
+#define MAX_TIMEOUT		(86400*31)	/* 31 days */
 
-/* will move it to <net/ip_vs.h> in the next version of ipvs */
-#ifndef IP_VS_SO_SET_RESTORE
-#define IP_VS_SO_SET_RESTORE     (IP_VS_BASE_CTL+13)
-#endif
-#ifndef IP_VS_SO_SET_SAVE
-#define IP_VS_SO_SET_SAVE        (IP_VS_BASE_CTL+14)
-#endif
-#define NUMBER_OF_CMD            (IP_VS_SO_SET_SAVE - IP_VS_BASE_CTL)
+#define NUMBER_OF_CMD		(IP_VS_SO_SET_MAX - IP_VS_BASE_CTL)
 
 static const char* cmdnames[] = {
 	"insert",
 	"add-service",
 	"edit-service",
-	"del-service",
+	"delele-service",
 	"flush",
 	"list",
 	"add-server",
-	"del-server",
+	"delete-server",
 	"edit-server",
 	"set",
 	"start-daemon",
@@ -170,7 +168,11 @@ static const char* cmdnames[] = {
 #define OPT_FORWARD	0x00080
 #define OPT_WEIGHT	0x00100
 #define OPT_MCAST	0x00200
-#define NUMBER_OF_OPT	10
+#define OPT_TIMEOUT	0x00400
+#define OPT_DAEMON	0x00800
+#define OPT_STATS	0x01000
+#define OPT_RATE	0x02000
+#define NUMBER_OF_OPT	14
 
 static const char* optnames[] = {
 	"numeric",
@@ -182,56 +184,59 @@ static const char* optnames[] = {
 	"real-server",
 	"forwarding-method",
 	"weight",
-	"mcast-interface"
+	"mcast-interface",
+	"timeout",
+	"daemon",
+	"stats",
+	"rate",
 };
 
 /*
  * Table of legal combinations of commands and options.
  * Key:
- *  +  compulsory
- *  x  illegal
- *     optional
+ *  '+'  compulsory
+ *  'x'  illegal
+ *  '1'  exclusive (only one '1' option can be supplied)
+ *  ' '  optional
  */
 static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 {
-	/*     -n   -c   svc  -s   -p   -M   -r   fwd  -w   mc */
-/*INSERT*/    {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
-/*ADD*/       {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x'},
-/*EDIT*/      {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x'},
-/*DEL*/       {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*LIST*/      {' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*ADD-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x'},
-/*DEL-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x'},
-/*EDIT-SRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x'},
-/*TIMEOUT*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*START-D*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' '},
-/*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
-/*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+	/*     -n   -c   svc  -s   -p   -M   -r   fwd  -w   -mc  -to  dmn  -st -rt */
+/*INSERT*/    {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'x', 'x', 'x', 'x'},
+/*ADD*/       {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDIT*/      {'x', 'x', '+', ' ', ' ', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*DEL*/       {'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*FLUSH*/     {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*LIST*/      {' ', '1', '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', '1', ' ', ' '},
+/*ADD-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
+/*DEL-SERVER*/{'x', 'x', '+', 'x', 'x', 'x', '+', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*EDIT-SRV*/  {'x', 'x', '+', 'x', 'x', 'x', '+', ' ', ' ', 'x', 'x', 'x', 'x', 'x'},
+/*TIMEOUT*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*START-D*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', ' ', 'x', 'x', 'x', 'x'},
+/*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 };
 
-
 /* printing format flags */
 #define FMT_NONE	0x0000
 #define FMT_NUMERIC	0x0001
 #define FMT_RULE	0x0002
+#define FMT_STATS	0x0004
+#define FMT_RATE	0x0008
 
 #define SERVICE_NONE    0x0
 #define SERVICE_ADDR    0x1
 #define SERVICE_PORT    0x2
 
-
 /* default scheduler */
 #define DEF_SCHED	"wlc"
 
 /* default multicast interface name */
 #define DEF_MCAST_IFN   "eth0"
 
-#define IPVS_PROC_FILE  "/proc/net/ip_vs"
 #define CONN_PROC_FILE  "/proc/net/ip_vs_conn"
 
-
 /* various parsing helpers & parsing functions */
 int str_is_digit(const char *str);
 int string_to_number(const char *s, int min, int max);
@@ -245,9 +250,9 @@ char * addrport_to_anyname(struct in_addr *addr, int port,
 			   unsigned short proto, unsigned int format);
 int parse_service(char *buf, u_int16_t proto,
 		  u_int32_t *addr, u_int16_t *port);
-int parse_fwmark(char *buf, u_int32_t *fwmark);
 int parse_netmask(char *buf, u_int32_t *addr);
-int parse_timeout(char *buf, unsigned *timeout);
+int parse_timeout(char *buf, int min, int max);
+unsigned int parse_fwmark(char *buf);
 
 /* check the options based on the commands_v_options table */
 static void generic_opt_check(int command, int options);
@@ -259,33 +264,46 @@ void usage_exit(const char *program, const int exit_status);
 void version_exit(int exit_status);
 void version(FILE *stream);
 void fail(int err, char *msg, ...);
-void check_ipvs_version(void);
-void list_vs(unsigned int format);
+
+/* various listing functions */
 void list_conn(unsigned int format);
 void list_service(struct ip_vs_rule_user *ur, unsigned int format);
-void print_vsinfo(char *buf, unsigned int format);
+void list_all(unsigned int format);
+void list_timeout(void);
+void list_daemon(void);
+
+int modprobe_ipvs(void);
+void check_ipvs_version(void);
 int process_options(int argc, char **argv, int reading_stdin);
 
 
 int main(int argc, char **argv)
 {
-	/*
-	 *      Warn the user if the IPVS version is out of date
-	 */
-	check_ipvs_version();
+	int result;
 
-	/*
-	 *      If no other arguement, list IPVS_PROC_FILE
-	 */
-	if (argc == 1){
-		list_vs(FMT_NONE);
-		exit(0);
+	if (ipvs_init()) {
+		/* try to insmod the ip_vs module if ipvs_init failed */
+		if (modprobe_ipvs() || ipvs_init())
+			fail(2, "Can't initialize ipvs: %s\n"
+				"Are you sure that IP Virtual Server is "
+				"built in the kernel or as module?",
+			     ipvs_strerror(errno));
 	}
 
-	/*
-	 *	Process command line arguments
-	 */
-	return process_options(argc, argv, 0);
+	/* warn the user if the IPVS version is out of date */
+	check_ipvs_version();
+
+	/* list the table if there is no other arguement */
+	if (argc == 1){
+		list_all(FMT_NONE);
+		return 0;
+	}
+
+	/* process command line arguments */
+	result = process_options(argc, argv, 0);
+
+	ipvs_close();
+	return result;
 }
 
 
@@ -330,6 +348,10 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"numeric", 'n', POPT_ARG_NONE, NULL, 'n'},
 		{"connection", 'c', POPT_ARG_NONE, NULL, 'c'},
 		{"mcast-interface", '\0', POPT_ARG_STRING, &optarg, '3'},
+		{"timeout", '\0', POPT_ARG_NONE, NULL, '5'},
+		{"daemon", '\0', POPT_ARG_NONE, NULL, '6'},
+		{"stats", '\0', POPT_ARG_NONE, NULL, '7'},
+		{"rate", '\0', POPT_ARG_NONE, NULL, '8'},
 		{NULL, 0, 0, NULL, 0}
 	};
 
@@ -416,10 +438,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			 * though it is not used
 			 */
 			ur->protocol = IPPROTO_TCP;
-			parse = parse_fwmark(optarg, &ur->vfwmark);
-			if (parse == 0 || ur->vfwmark == 0)
-				fail(2, "illegal virtual server "
-				     "fwmark specified");
+			ur->vfwmark = parse_fwmark(optarg);
 			break;
 		case 's':
 			set_option(options, OPT_SCHEDULER);
@@ -428,15 +447,11 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		case 'p':
 			set_option(options, OPT_PERSISTENT);
 			ur->vs_flags = IP_VS_SVC_F_PERSISTENT;
-			parse = parse_timeout(optarg, &ur->timeout);
-			if (parse == 0)
-				fail(2, "illegal timeout for "
-				     "persistent service");
+			ur->timeout = parse_timeout(optarg, 1, MAX_TIMEOUT);
 			break;
 		case 'M':
 			set_option(options, OPT_NETMASK);
-			parse = parse_netmask(optarg,
-					      &ur->netmask);
+			parse = parse_netmask(optarg, &ur->netmask);
 			if (parse != 1)
 				fail(2, "illegal virtual server "
 				     "persistent mask specified");
@@ -469,7 +484,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		case 'w':
 			set_option(options, OPT_WEIGHT);
 			if ((ur->weight =
-			     string_to_number(optarg,0,65535)) == -1)
+			     string_to_number(optarg, 0, 65535)) == -1)
 				fail(2, "illegal weight specified");
 			break;
 		case 'c':
@@ -483,6 +498,20 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			set_option(options, OPT_MCAST);
 			strncpy(ur->mcast_ifn, optarg, IP_VS_IFNAME_MAXLEN);
 			break;
+		case '5':
+			set_option(options, OPT_TIMEOUT);
+			break;
+		case '6':
+			set_option(options, OPT_DAEMON);
+			break;
+		case '7':
+			set_option(options, OPT_STATS);
+			*format |= FMT_STATS;
+			break;
+		case '8':
+			set_option(options, OPT_RATE);
+			*format |= FMT_RATE;
+			break;
 		default:
 			fail(2, "invalid option");
 		}
@@ -503,11 +532,12 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		if ((optarg=(char *)poptGetArg(context))
 		    && (optarg1=(char *)poptGetArg(context))
 		    && (optarg2=(char *)poptGetArg(context))) {
-			parse = parse_timeout(optarg, &ur->tcp_timeout)
-				&& parse_timeout(optarg1, &ur->tcp_fin_timeout)
-				&& parse_timeout(optarg2, &ur->udp_timeout);
-			if (parse == 0)
-				fail(2, "illegal timeout values");
+			ur->tcp_timeout = parse_timeout(optarg,
+							0, MAX_TIMEOUT);
+			ur->tcp_fin_timeout = parse_timeout(optarg1,
+							    0, MAX_TIMEOUT);
+			ur->udp_timeout = parse_timeout(optarg2,
+							0, MAX_TIMEOUT);
 		} else
 			fail(2, "--set option requires 3 timeout values");
 	}
@@ -558,6 +588,10 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"numeric", 0, 0, 'n'},
 		{"connection", 0, 0, 'c'},
 		{"mcast-interface", 1, 0, '3'},
+		{"timeout", 0, 0, '5'},
+		{"daemon", 0, 0, '6'},
+		{"stats", 0, 0, '7'},
+		{"rate", 0, 0, '8'},
 		{"help", 0, 0, 'h'},
 		{0, 0, 0, 0}
 	};
@@ -607,13 +641,12 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		    && argv[optind][0] != '-'
 		    && argv[optind+1][0] != '!'
 		    && argv[optind+1][0] != '-') {
-			parse = parse_timeout(optarg, &ur->tcp_timeout)
-				&& parse_timeout(argv[optind++],
-						 &ur->tcp_fin_timeout)
-				&& parse_timeout(argv[optind++],
-						 &ur->udp_timeout);
-			if (parse == 0)
-				fail(2, "illegal timeout values");
+			ur->tcp_timeout = parse_timeout(optarg,
+							0, MAX_TIMEOUT);
+			ur->tcp_fin_timeout =  parse_timeout(argv[optind++],
+							     0, MAX_TIMEOUT);
+			ur->udp_timeout = parse_timeout(argv[optind++],
+							0, MAX_TIMEOUT);
 		} else
 			fail(2, "-s option requires 3 timeout values");
 		break;
@@ -666,10 +699,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			 * though it is not used
 			 */
 			ur->protocol = IPPROTO_TCP;
-			parse = parse_fwmark(optarg, &ur->vfwmark);
-			if (parse == 0 || ur->vfwmark == 0)
-				fail(2, "illegal virtual server "
-				     "fwmark specified");
+			ur->vfwmark = parse_fwmark(optarg);
 			break;
 		case 's':
 			set_option(options, OPT_SCHEDULER);
@@ -681,16 +711,11 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			if (!optarg && optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!')
 				optarg = argv[optind++];
-			parse = parse_timeout(optarg,
-					      &ur->timeout);
-			if (parse == 0)
-				fail(2, "illegal timeout "
-				     "for persistent service");
+			ur->timeout = parse_timeout(optarg, 1, MAX_TIMEOUT);
 			break;
 		case 'M':
 			set_option(options, OPT_NETMASK);
-			parse = parse_netmask(optarg,
-					      &ur->netmask);
+			parse = parse_netmask(optarg, &ur->netmask);
 			if (parse != 1)
 				fail(2, "illegal virtual server "
 				     "persistent mask specified");
@@ -737,6 +762,20 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 			set_option(options, OPT_MCAST);
 			strncpy(ur->mcast_ifn, optarg, IP_VS_IFNAME_MAXLEN);
 			break;
+		case '5':
+			set_option(options, OPT_TIMEOUT);
+			break;
+		case '6':
+			set_option(options, OPT_DAEMON);
+			break;
+		case '7':
+			set_option(options, OPT_STATS);
+			*format |= FMT_STATS;
+			break;
+		case '8':
+			set_option(options, OPT_RATE);
+			*format |= FMT_RATE;
+			break;
 		default:
 			fail(2, "invalid option");
 		}
@@ -751,13 +790,32 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 #endif /* HAVE_POPT */
 
 
+int restore_table(int argc, char **argv, int reading_stdin)
+{
+	int result=0;
+	dynamic_array_t *a;
+
+	/* avoid infinite loop */
+	if (reading_stdin != 0)
+		tryhelp_exit(argv[0], -1);
+
+	while ((a = config_stream_read(stdin, argv[0])) != NULL) {
+		int i;
+		if ((i = (int)dynamic_array_get_count(a)) > 1)
+			result = process_options
+				(i, (char **)dynamic_array_get_vector(a), 1);
+		dynamic_array_destroy(a, DESTROY_STR);
+	}
+	return result;
+}
+
 int process_options(int argc, char **argv, int reading_stdin)
 {
 	struct ip_vs_rule_user urule;
 	int cmd=IP_VS_SO_SET_NONE;
 	unsigned int options=OPT_NONE;
 	unsigned int format=FMT_NONE;
-	int sockfd, result;
+	int result;
 
 	memset(&urule, 0, sizeof(struct ip_vs_rule_user));
 	/* Set the default weight 1 */
@@ -772,37 +830,30 @@ int process_options(int argc, char **argv, int reading_stdin)
 
 	generic_opt_check(cmd, options);
 
-	if (cmd == IP_VS_SO_SET_LIST) {
-		if (options & OPT_CONNECTION) {
-			if (options & OPT_SERVICE)
-				fail(2, "Can't use -c with service-address");
+	switch (cmd) {
+	case IP_VS_SO_SET_LIST:
+		if ((options & OPT_CONNECTION
+		     || options & OPT_TIMEOUT
+		     || options & OPT_DAEMON)
+		    && (options & OPT_STATS || options & OPT_RATE))
+			fail(2, "options conflicts in the list command");
+
+		if (options & OPT_CONNECTION)
 			list_conn(format);
-		} else if (options & OPT_SERVICE)
+		else if (options & OPT_SERVICE)
 			list_service(&urule, format);
+		else if (options & OPT_TIMEOUT)
+			list_timeout();
+		else if (options & OPT_DAEMON)
+			list_daemon();
 		else
-			list_vs(format);
+			list_all(format);
 		return 0;
-	} else if (cmd == IP_VS_SO_SET_RESTORE) {
-		int result=0;
-		dynamic_array_t *a;
-
-		/* avoid infinite loop */
-		if (reading_stdin != 0)
-			tryhelp_exit(argv[0], -1);
-
-		while ((a = config_stream_read(stdin, argv[0])) != NULL) {
-			int i;
-			if ((i = (int)dynamic_array_get_count(a)) > 1)
-				result = process_options
-					(i,
-					 (char **)dynamic_array_get_vector(a),
-					 1);
-			dynamic_array_destroy(a, DESTROY_STR);
-		}
-		return result;
-	} else if (cmd == IP_VS_SO_SET_SAVE) {
+	case IP_VS_SO_SET_RESTORE:
+		return restore_table(argc, argv, reading_stdin);
+	case IP_VS_SO_SET_SAVE:
 		format |= FMT_RULE;
-		list_vs(format);
+		list_all(format);
 		return 0;
 	}
 
@@ -822,8 +873,7 @@ int process_options(int argc, char **argv, int reading_stdin)
 			strcpy(urule.sched_name, DEF_SCHED);
 	}
 
-	if (cmd == IP_VS_SO_SET_STARTDAEMON
-	    && strlen(urule.mcast_ifn) == 0)
+	if (cmd == IP_VS_SO_SET_STARTDAEMON && strlen(urule.mcast_ifn) == 0)
 		strcpy(urule.mcast_ifn, DEF_MCAST_IFN);
 
 	if (cmd == IP_VS_SO_SET_ADDDEST || cmd == IP_VS_SO_SET_EDITDEST) {
@@ -838,77 +888,63 @@ int process_options(int argc, char **argv, int reading_stdin)
 			urule.dport = urule.vport;
 	}
 
-	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
-	if (sockfd == -1) {
-		perror("socket creation failed");
-		exit(1);
-	}
+	result = ipvs_command(cmd, &urule);
+	if (result)
+		fprintf(stderr, "%s\n", ipvs_strerror(errno));
 
-	result = setsockopt(sockfd, IPPROTO_IP, cmd,
-			    (char *)&urule, sizeof(urule));
-	if (result) {
-		perror("setsockopt failed");
-
-		/*
-		 *    Print most common error messages
-		 */
-		switch (cmd) {
-		case IP_VS_SO_SET_ADD:
-			if (errno == EEXIST)
-				printf("Service already exists\n");
-			else if (errno == ENOENT)
-				printf("Scheduler not found: ip_vs_%s.o\n",
-				       urule.sched_name);
-			break;
-		case IP_VS_SO_SET_EDIT:
-			if (errno==ESRCH)
-				printf("No such service\n");
-			else if (errno == ENOENT)
-				printf("Scheduler not found: ip_vs_%s.o\n",
-				       urule.sched_name);
-			break;
-		case IP_VS_SO_SET_DEL:
-			if (errno==ESRCH)
-				printf("No such service\n");
-			break;
-		case IP_VS_SO_SET_ADDDEST:
-			if (errno == ESRCH)
-				printf("Service not defined\n");
-			else if (errno == EEXIST)
-				printf("Destination already exists\n");
-			break;
-		case IP_VS_SO_SET_EDITDEST:
-		case IP_VS_SO_SET_DELDEST:
-			if (errno==ESRCH)
-				printf("Service not defined\n");
-			else if (errno == ENOENT)
-				printf("No such destination\n");
-			break;
-		}
-	}
-
-	close(sockfd);
 	return result;
 }
 
 
 int string_to_number(const char *s, int min, int max)
 {
-	int number;
+	long number;
 	char *end;
 
-	number = (int)strtol(s, &end, 10);
+	errno = 0;
+	number = strtol(s, &end, 10);
 	if (*end == '\0' && end != s) {
-		/*
-		 * We parsed a number, let's see if we want this.
-		 * If max <= min then ignore ranges
-		 */
-		if (max <= min || ( min <= number && number <= max))
+		/* We parsed a number, let's see if we want this. */
+		if (errno != ERANGE && min <= number && number <= max)
 			return number;
-		else
-			return -1;
-	} else
-		return -1;
+	}
+	return -1;
+}
+
+
+/*
+ * Parse the timeout value.
+ */
+int parse_timeout(char *buf, int min, int max)
+{
+	int i;
+
+	/* it is just for parsing timeout of persistent service */
+	if (buf == NULL)
+		return IP_VS_TEMPLATE_TIMEOUT;
+
+	if ((i=string_to_number(buf, min, max)) == -1)
+		fail(2, "invalid timeout value `%s' specified", buf);
+
+	return i * HZ;
+}
+
+
+/*
+ * Parse IP fwmark from the argument.
+ */
+unsigned int parse_fwmark(char *buf)
+{
+	unsigned long l;
+	char *end;
+
+	errno = 0;
+	l = strtol(buf, &end, 10);
+	if (*end != '\0' || end == buf ||
+	    errno == ERANGE || l <= 0 || l > UINT_MAX)
+		fail(2, "invalid fwmark value `%s' specified", buf);
+
+	return l;
 }
 
 
@@ -934,23 +970,6 @@ int parse_netmask(char *buf, u_int32_t *addr)
 	return 1;
 }
 
-/*
- * Get IP fwmark from the argument.
- * Result is
- * 1 on success
- * 0 on error
- */
-int parse_fwmark(char *buf, u_int32_t *fwmark)
-{
-	long tmpl;
-
-	if ((tmpl=string_to_number(buf, 0, 0)) == -1)
-		return 0;
-	*fwmark = tmpl;
-
-	return 1;
-}
-
 
 /*
  * Get IP address and port from the argument.
@@ -997,32 +1016,11 @@ int parse_service(char *buf, u_int16_t proto, u_int32_t *addr, u_int16_t *port)
 }
 
 
-/*
- * Get the timeout of persistent service.
- * Return 0 if failed(the timeout value is less or equal than zero).
- *	  1 if succeed.
- */
-int parse_timeout(char *buf, unsigned *timeout)
-{
-	int i;
-
-	if (buf == NULL) {
-		*timeout = IP_VS_TEMPLATE_TIMEOUT;
-		return 1;
-	}
-
-	if ((i=string_to_number(buf, 1, 86400*31)) == -1)
-		return 0;
-
-	*timeout = i * HZ;
-	return 1;
-}
-
-
 static void
 generic_opt_check(int command, int options)
 {
 	int i, j;
+	int last = 0, count = 0;
 
 	/* Check that commands are valid with options. */
 	i = command - IP_VS_BASE_CTL -1;
@@ -1038,6 +1036,16 @@ generic_opt_check(int command, int options)
 				fail(2, "Illegal '%s' option with "
 				     "the '%s' command",
 				     optnames[j], cmdnames[i]);
+			if (commands_v_options[i][j] == '1') {
+				count++;
+				if (count == 1) {
+					last = j;
+					continue;
+				}
+				fail(2, "The option '%s' conflicts with the "
+				     "'%s' option in the '%s' command",
+				     optnames[j], optnames[last], cmdnames[i]);
+			}
 		}
 	}
 }
@@ -1139,6 +1147,10 @@ void usage_exit(const char *program, const int exit_status)
 		"  --weight       -w weight            capacity of real server\n"
 		"  --mcast-interface interface         multicast interface for connection sync\n"
 		"  --connection   -c                   output of current IPVS connections\n"
+		"  --timeout                           output of timeout (tcp tcpfin udp)\n"
+		"  --daemon                            output of daemon information\n"
+		"  --stats                             output of statistics information\n"
+		"  --rate                              output of rate information\n"
 		"  --numeric      -n                   numeric output of addresses and ports\n",
 		DEF_SCHED);
 
@@ -1175,79 +1187,52 @@ void fail(int err, char *msg, ...)
 	va_list args;
 
 	va_start(args, msg);
-	vprintf(msg, args);
+	vfprintf(stderr, msg, args);
 	va_end(args);
-	printf("\n");
+	fprintf(stderr, "\n");
 	exit(err);
 }
 
 
 int modprobe_ipvs(void)
 {
-    char *argv[] = { "/sbin/modprobe", "-s", "-k", "--", "ip_vs", NULL };
-    int child;
-    int status;
-    int rc;
+	char *argv[] = { "/sbin/modprobe", "-s", "-k", "--", "ip_vs", NULL };
+	int child;
+	int status;
+	int rc;
 
-    if (!(child = fork())) {
-	    execv(argv[0], argv);
-	    exit(1);
-    }
+	if (!(child = fork())) {
+		execv(argv[0], argv);
+		exit(1);
+	}
 
-    rc = waitpid(child, &status, 0);
+	rc = waitpid(child, &status, 0);
 
-    if (!WIFEXITED(status) || WEXITSTATUS(status)) {
-	    return 1;
-    }
+	if (!WIFEXITED(status) || WEXITSTATUS(status)) {
+		return 1;
+	}
 
-    return 0;
+	return 0;
 }
 
 
 void check_ipvs_version(void)
 {
-	static char buffer[128];
-	int major;
-	int minor;
-	int patch;
-	FILE *handle;
-
-	handle = fopen(IPVS_PROC_FILE, "r");
-	if (!handle) {
-		/* try to request the ip_vs module */
-		if (modprobe_ipvs()
-		    || !(handle = fopen(IPVS_PROC_FILE, "r"))) {
-			fprintf(stderr, "Could not open the %s file\n"
-				"Are you sure that IP Virtual Server is "
-				"built in the kernel or as module?\n",
-				IPVS_PROC_FILE);
-			exit(1);
-		}
+	/* verify the IPVS version */
+	if (ipvs_info.version <
+	    IPVS_VERSION(MINIMUM_IPVS_VERSION_MAJOR,
+			 MINIMUM_IPVS_VERSION_MINOR,
+			 MINIMUM_IPVS_VERSION_PATCH)) {
+		fprintf(stderr,
+			"Warning: IPVS version missmatch: \n"
+			"  Kernel compiled with IPVS version %d.%d.%d\n"
+			"  ipvsadm " IPVSADM_VERSION_NO
+			" requires minimum IPVS version %d.%d.%d\n\n",
+			NVERSION(ipvs_info.version),
+			MINIMUM_IPVS_VERSION_MAJOR,
+			MINIMUM_IPVS_VERSION_MINOR,
+			MINIMUM_IPVS_VERSION_PATCH);
 	}
-
-	/*
-	 * Read the first line and verify the IPVS version
-	 */
-	if (!feof(handle) && fgets(buffer, sizeof(buffer), handle)) {
-		sscanf(buffer, "%*128[a-z A-Z] %d.%d.%d",
-		       &major, &minor, &patch);
-		if (IPVS_VERSION(major,minor,patch) <
-		    IPVS_VERSION(MINIMUM_IPVS_VERSION_MAJOR,
-				 MINIMUM_IPVS_VERSION_MINOR,
-				 MINIMUM_IPVS_VERSION_PATCH)) {
-			fprintf(stderr,
-				"Warning: IPVS version missmatch: \n"
-				"  Kernel compiled with IPVS version %d.%d.%d\n"
-				"  ipvsadm " IPVSADM_VERSION_NO
-				" requires minimum IPVS version %d.%d.%d\n\n",
-				major, minor, patch,
-				MINIMUM_IPVS_VERSION_MAJOR,
-				MINIMUM_IPVS_VERSION_MINOR,
-				MINIMUM_IPVS_VERSION_PATCH);
-		}
-		buffer[strlen(buffer)-1] = '\0';
-	}
-	fclose(handle);
 }
 
 
@@ -1341,9 +1326,12 @@ void list_conn(unsigned int format)
 
 static inline char *fwd_name(unsigned flags)
 {
-	char *fwd;
+	char *fwd = NULL;
 
 	switch (flags & IP_VS_CONN_F_FWD_MASK) {
+	case IP_VS_CONN_F_MASQ:
+		fwd = "Masq";
+		break;
 	case IP_VS_CONN_F_LOCALNODE:
 		fwd = "Local";
 		break;
@@ -1353,277 +1341,239 @@ static inline char *fwd_name(unsigned flags)
 	case IP_VS_CONN_F_DROUTE:
 		fwd = "Route";
 		break;
-	default:
-		fwd = "Masq";
 	}
 	return fwd;
 }
 
-void list_service(struct ip_vs_rule_user *ur, unsigned int format)
+static inline char *fwd_switch(unsigned flags)
+{
+	char *swt = NULL;
+
+	switch (flags & IP_VS_CONN_F_FWD_MASK) {
+	case IP_VS_CONN_F_MASQ:
+		swt = "-m"; break;
+	case IP_VS_CONN_F_TUNNEL:
+		swt = "-i"; break;
+	case IP_VS_CONN_F_LOCALNODE:
+	case IP_VS_CONN_F_DROUTE:
+		swt = "-g"; break;
+	}
+	return swt;
+}
+
+
+static void print_largenum(unsigned long long i)
+{
+	if (i < 100000000)			/* less than 100 million */
+		printf("%9llu", i);
+	else if (i < 1000000000)		/* less than 1 billion */
+		printf("%8lluK", i / 1000);
+	else if (i < 100000000000ULL)		/* less than 100 billion */
+		printf("%8lluM", i / 1000000);
+	else if (i < 100000000000000ULL)	/* less than 100 trillion */
+		printf("%8lluG", i / 1000000000ULL);
+	else
+		printf("%8lluT", i / 1000000000000ULL);
+}
+
+
+static void print_title(unsigned int format)
+{
+	if (format & FMT_STATS)
+		printf("%-33s %8s %8s %8s %8s %8s\n"
+		       "  -> RemoteAddress:Port\n",
+		       "Prot LocalAddress:Port",
+		       "Conns", "InPkts", "OutPkts", "InBytes", "OutBytes");
+	else if (format & FMT_RATE)
+		printf("%-33s %8s %8s %8s %8s %8s\n"
+		       "  -> RemoteAddress:Port\n",
+		       "Prot LocalAddress:Port",
+		       "CPS", "InPPS", "OutPPS", "InBPS", "OutBPS");
+	else if (!(format & FMT_RULE))
+		printf("Prot LocalAddress:Port Scheduler Flags\n"
+		       "  -> RemoteAddress:Port           Forward Weight ActiveConn InActConn\n");
+}
+
+
+static void
+print_service(struct ip_vs_service_user *svc, unsigned int format)
 {
-	struct ip_vs_get_service svc;
 	struct ip_vs_get_dests *d;
-	socklen_t len;
-	int sockfd;
+	char svc_name[64];
 	int i;
-	struct in_addr vaddr;
-	struct in_addr daddr;
-	char *vname;
-	char *dname;
 
-	memset(&svc, 0, sizeof(svc));
-	svc.fwmark = ur->vfwmark;
-	svc.protocol = ur->protocol;
-	svc.addr = ur->vaddr;
-	svc.port = ur->vport;
-	len = sizeof(svc);
-
-	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
-	if (sockfd == -1) {
-		perror("socket creation failed");
+	if (!(d = ipvs_get_dests(svc)))
 		exit(1);
-	}
 
-	if (getsockopt(sockfd, IPPROTO_IP, IP_VS_SO_GET_SERVICE,
-		       (char *)&svc, &len)) {
-		perror("getsockopt failed");
-		exit(1);
-	}
+	if (svc->fwmark) {
+		if (format & FMT_RULE)
+			sprintf(svc_name, "-f %d", svc->fwmark);
+		else
+			sprintf(svc_name, "FWM  %d", svc->fwmark);
+	} else {
+		struct in_addr vaddr;
+		char *vname;
+		vaddr.s_addr = svc->addr;
 
-	len = sizeof(struct ip_vs_get_dests) +
-		sizeof(struct ip_vs_dest_entry)*svc.num_dests;
-	if (!(d = malloc(len))) {
-		perror("no memory");
-		exit(1);
+		if (!(vname = addrport_to_anyname(&vaddr, ntohs(svc->port),
+						  svc->protocol, format)))
+			fail(2, "addrport_to_anyname: %s", strerror(errno));
+		if (format & FMT_RULE)
+			sprintf(svc_name, "%s %s",
+				svc->protocol==IPPROTO_TCP?"-t":"-u",
+				vname);
+		else
+			sprintf(svc_name, "%s  %s",
+				svc->protocol==IPPROTO_TCP?"TCP":"UDP",
+				vname);
+		free(vname);
+		svc_name[33] = '\0';
 	}
-	d->fwmark = svc.fwmark;
-	d->protocol = svc.protocol;
-	d->addr = svc.addr;
-	d->port = svc.port;
-	d->num_dests = svc.num_dests;
-
-	if (getsockopt(sockfd, IPPROTO_IP, IP_VS_SO_GET_DESTS, d, &len) < 0) {
-		perror("getsockopt failed");
-		free(d);
-		exit(1);
-	}
-
-	printf("Prot LocalAddress:Port Scheduler Flags\n"
-	       "  -> RemoteAddress:Port             Forward Weight ActiveConn InActConn\n");
 
 	/* print virtual service info */
-	if (svc.fwmark)
-		printf("FWM  %d %s", svc.fwmark, svc.sched_name);
-	else {
-		vaddr.s_addr = svc.addr;
-		if (!(vname = addrport_to_anyname(&vaddr, ntohs(svc.port),
-						  svc.protocol, format)))
-			exit(1);
-		printf("%s  %s %s", svc.protocol==IPPROTO_TCP?"TCP":"UDP",
-		       vname, svc.sched_name);
-		free(vname);
-	}
-	if (svc.flags & IP_VS_SVC_F_PERSISTENT) {
-		printf(" persistent %u", svc.timeout/HZ);
-		if (svc.netmask != (unsigned long int) 0xffffffff) {
-			struct in_addr mask;
-			mask.s_addr = svc.netmask;
-			printf(" mask %s", inet_ntoa(mask));
+	if (format & FMT_RULE) {
+		printf("-A %s -s %s", svc_name, svc->sched_name);
+		if (svc->flags & IP_VS_SVC_F_PERSISTENT) {
+			printf(" -p %u", svc->timeout/HZ);
+			if (svc->netmask != (unsigned long int) 0xffffffff) {
+				struct in_addr mask;
+				mask.s_addr = svc->netmask;
+				printf(" -M %s", inet_ntoa(mask));
+			}
+		}
+	} else if (format & FMT_STATS) {
+		printf("%-33s", svc_name);
+		print_largenum(svc->stats.conns);
+		print_largenum(svc->stats.inpkts);
+		print_largenum(svc->stats.outpkts);
+		print_largenum(svc->stats.inbytes);
+		print_largenum(svc->stats.outbytes);
+	} else if (format & FMT_RATE) {
+		printf("%-33s", svc_name);
+		print_largenum(svc->stats.cps);
+		print_largenum(svc->stats.inpps);
+		print_largenum(svc->stats.outpps);
+		print_largenum(svc->stats.inbps);
+		print_largenum(svc->stats.outbps);
+	} else {
+		printf("%s %s", svc_name, svc->sched_name);
+		if (svc->flags & IP_VS_SVC_F_PERSISTENT) {
+			printf(" persistent %u", svc->timeout/HZ);
+			if (svc->netmask != (unsigned long int) 0xffffffff) {
+				struct in_addr mask;
+				mask.s_addr = svc->netmask;
+				printf(" mask %s", inet_ntoa(mask));
+			}
 		}
 	}
 	printf("\n");
 
 	/* print all the destination entries */
 	for (i=0; i<d->num_dests; i++) {
-		struct ip_vs_dest_entry *e = &d->entrytable[i];
+		struct in_addr daddr;
+		char *dname;
+		struct ip_vs_dest_user *e = &d->entrytable[i];
 
 		daddr.s_addr = e->addr;
 		if (!(dname = addrport_to_anyname(&daddr, ntohs(e->port),
-						  svc.protocol, format)))
+						  svc->protocol, format)))
 			exit(1);
-		printf("  -> %-30s %-7s %-6d %-10u %-10u\n",
-		       dname, fwd_name(e->flags),
-		       e->weight, e->activeconns, e->inactconns);
+		dname[28] = '\0';
+		if (format & FMT_RULE) {
+			printf("-a %s -r %s %s -w %d\n", svc_name, dname,
+			       fwd_switch(e->flags), e->weight);
+		} else if (format & FMT_STATS) {
+			printf("  -> %-28s", dname);
+			print_largenum(e->stats.conns);
+			print_largenum(e->stats.inpkts);
+			print_largenum(e->stats.outpkts);
+			print_largenum(e->stats.inbytes);
+			print_largenum(e->stats.outbytes);
+			printf("\n");
+		} else if (format & FMT_RATE) {
+			printf("  -> %-28s %8u %8u %8u", dname,
+			       e->stats.cps,
+			       e->stats.inpps,
+			       e->stats.outpps);
+			print_largenum(e->stats.inbps);
+			print_largenum(e->stats.outbps);
+			printf("\n");
+		} else
+			printf("  -> %-28s %-7s %-6d %-10u %-10u\n",
+			       dname, fwd_name(e->flags),
+			       e->weight, e->activeconns, e->inactconns);
 		free(dname);
 	}
+	free(d);
 }
 
 
-void list_vs(unsigned int format)
+void list_service(struct ip_vs_rule_user *ur, unsigned int format)
 {
-	static char buffer[256];
-	FILE *handle;
+	struct ip_vs_service_user *svc;
+
+	if (!(svc = ipvs_get_service(ur->vfwmark, ur->protocol,
+				     ur->vaddr, ur->vport)))
+		exit(1);
+
+	print_title(format);
+	print_service(svc, format);
+	free(svc);
+}
+
+
+void list_all(unsigned int format)
+{
+	struct ip_vs_get_services *get;
 	int i;
 
-	handle = fopen(IPVS_PROC_FILE, "r");
-	if (!handle) {
-		fprintf(stderr, "Could not open the %s file\n"
-			"Are you sure that IP Virtual Server is supported "
-			"by the kernel?\n", IPVS_PROC_FILE);
+	if (!(format & FMT_RULE))
+		printf("IP Virtual Server version %d.%d.%d (size=%d)\n",
+		       NVERSION(ipvs_info.version), ipvs_info.size);
+
+	if (!(get = ipvs_get_services()))
 		exit(1);
-	}
 
-	/*
-	 * Read and print the first three head lines
-	 */
-	for (i=0; i<2 && !feof(handle); i++) {
-		if (fgets(buffer, sizeof(buffer), handle)
-		    && !(format & FMT_RULE))
-			printf("%s", buffer);
-	}
-	if (fgets(buffer, sizeof(buffer), handle) && !(format & FMT_RULE))
-		printf("  -> RemoteAddress:Port             "
-		       "Forward Weight ActiveConn InActConn\n");
-
-	/*
-	 * Print the VS information according to the format
-	 */
-	while (!feof(handle)) {
-		if (fgets(buffer, sizeof(buffer), handle))
-			print_vsinfo(buffer, format);
-	}
-
-	fclose(handle);
+	print_title(format);
+	for (i=0; i<get->num_services; i++)
+		print_service(&get->entrytable[i], format);
+	free(get);
 }
 
 
-char *get_fwd_switch(char *fwd)
+void list_timeout(void)
 {
-	char *swt;
+	struct ip_vs_timeout_user *u;
 
-	switch (fwd[0]) {
-	case 'M':
-		swt = "-m"; break;
-	case 'T':
-		swt = "-i"; break;
+	if (!(u = ipvs_get_timeouts()))
+		exit(1);
+	printf("Timeout (tcp tcpfin udp): %d %d %d\n",
+	       u->tcp_timeout/HZ, u->tcp_fin_timeout/HZ, u->udp_timeout/HZ);
+	free(u);
+}
+
+
+void list_daemon(void)
+{
+	struct ip_vs_daemon_user *u;
+
+	if (!(u = ipvs_get_daemon()))
+		exit(1);
+
+	switch (u->state) {
+	case IP_VS_STATE_NONE:
+		printf("not running\n");
+		break;
+	case IP_VS_STATE_MASTER:
+		printf("running as master (mcast=%s)\n", u->mcast_ifn);
+		break;
+	case IP_VS_STATE_BACKUP:
+		printf("running as backup (mcast=%s)\n", u->mcast_ifn);
+		break;
 	default:
-		swt = "-g"; break;
-	}
-	return swt;
-}
-
-void print_vsinfo(char *buf, unsigned int format)
-{
-	/* virtual service variables */
-	char protocol[10];
-	static unsigned short  proto = 0;
-	static struct in_addr  vaddr;
-	static unsigned short  vport;
-	static u_int32_t fwmark;
-	char scheduler[10];
-	char flags[40];
-	unsigned int timeout;
-	struct in_addr  vmask;
-
-	/* destination variables */
-	static char arrow[10];
-	struct in_addr	daddr;
-	unsigned short	dport;
-	static char fwd[10];
-	int weight;
-	int activeconns;
-	int inactconns;
-
-	int n;
-	unsigned long temp;
-	unsigned long temp2;
-	char *vname;
-	char *dname;
-
-	if (buf[0] == ' ') {
-		/* destination entry */
-		if ((n = sscanf(buf, " %s %lX:%hX %s %d %d %d",
-				arrow, &temp, &dport, fwd, &weight,
-				&activeconns, &inactconns)) == -1)
-			exit(1);
-		if (n != 7)
-			fail(2, "unexpected input data");
-
-		daddr.s_addr = (__u32) htonl(temp);
-
-		if (!(dname=addrport_to_anyname(&daddr, dport, proto, format)))
-			exit(1);
-		if (format & FMT_RULE) {
-			if (fwmark == 0) {
-				if (!(vname=addrport_to_anyname
-				      (&vaddr, vport, proto, format)))
-					exit(1);
-
-				printf("-a %s %s -r %s %s -w %d\n",
-				       proto==IPPROTO_TCP?"-t":"-u",
-				       vname, dname, get_fwd_switch(fwd), weight);
-				free(vname);
-			} else {
-				printf("-a -f %d -r %s %s -w %d\n", fwmark,
-				       dname, get_fwd_switch(fwd), weight);
-			}
-		} else {
-			printf("  -> %-30s %-7s %-6d %-10d %-10d\n",
-			       dname , fwd, weight, activeconns, inactconns);
-		}
-		free(dname);
-	} else if (buf[0] == 'F') {
-		/* fwmark virtual service entry */
-		if ((n = sscanf(buf, "%s %X %s %s %u %lX", protocol,
-				&fwmark, scheduler, flags,
-				&timeout, &temp2)) == -1)
-			exit(1);
-		if (n!=6 && n!=3)
-			fail(2, "unexpected input data");
-
-		vmask.s_addr = (__u32) htonl(temp2);
-
-		if (format & FMT_RULE)
-			printf("-A -f %d -s %s", fwmark, scheduler);
-		else
-			printf("%s  %d %s", protocol, fwmark, scheduler);
-
-		if (n > 3) {
-			printf(" %s %d",
-			       format&FMT_RULE?"-p":flags, timeout/HZ);
-			if (vmask.s_addr != (unsigned long int) 0xffffffff)
-				printf(" %s %s",
-				       format&FMT_RULE?"-M":"mask",
-				       inet_ntoa(vmask));
-		}
-		printf("\n");
-	} else {
-		/* TCP/UDP virtual service entry  */
-		fwmark=0;  /* Reset firewall mark to unused */
-
-		if ((n = sscanf(buf, "%s %lX:%hX %s %s %u %lX",
-				protocol, &temp, &vport, scheduler,
-				flags, &timeout, &temp2)) == -1)
-			exit(1);
-		if (n!=7 && n!=4)
-			fail(2, "unexpected input data");
-
-		vaddr.s_addr = (__u32) htonl(temp);
-		vmask.s_addr = (__u32) htonl(temp2);
-		if (strcmp(protocol, "TCP") == 0) proto = IPPROTO_TCP;
-		else if (strcmp(protocol, "UDP") == 0) proto = IPPROTO_UDP;
-		else proto = 0;
-
-		if (!(vname=addrport_to_anyname(&vaddr, vport, proto, format)))
-			exit(1);
-
-		if (format & FMT_RULE)
-			printf("-A %s %s -s %s", proto==IPPROTO_TCP?"-t":"-u",
-			       vname, scheduler);
-		else
-			printf("%s  %s %s", protocol, vname, scheduler);
-
-		if (n > 4) {
-			printf(" %s %d",
-			       format&FMT_RULE?"-p":flags, timeout/HZ);
-			if (vmask.s_addr != (unsigned long int) 0xffffffff)
-				printf(" %s %s\n",
-				       format&FMT_RULE?"-M":"mask",
-				       inet_ntoa(vmask));
-		}
-		printf("\n");
-		free(vname);
+		fail(2, "invalid state");
 	}
+	free(u);
 }
 
 
